@@ -21,6 +21,9 @@ import types
 import unittest
 import weakref
 import os
+import enum
+import typing
+from functools import wraps
 
 import pytest
 
@@ -43,9 +46,11 @@ import cloudpickle
 from cloudpickle.cloudpickle import _is_dynamic
 from cloudpickle.cloudpickle import _make_empty_cell, cell_set
 from cloudpickle.cloudpickle import _extract_class_dict, _whichmodule
+from cloudpickle.cloudpickle import _lookup_module_and_qualname
 
 from .testutils import subprocess_pickle_echo
 from .testutils import assert_run_python_script
+from .testutils import subprocess_worker
 
 
 _TEST_GLOBAL_VARIABLE = "default_value"
@@ -181,11 +186,7 @@ class CloudPickleTest(unittest.TestCase):
         def foo():
             sys.exit(0)
 
-        func_code = getattr(foo, '__code__', None)
-        if func_code is None:  # PY2 backwards compatibility
-            func_code = foo.func_code
-
-        self.assertTrue("exit" in func_code.co_names)
+        self.assertTrue("exit" in foo.__code__.co_names)
         cloudpickle.dumps(foo)
 
     def test_buffer(self):
@@ -204,9 +205,6 @@ class CloudPickleTest(unittest.TestCase):
         self.assertEqual(pickle_depickle(buffer_obj, protocol=self.protocol),
                          buffer_obj.tobytes())
 
-    @pytest.mark.skipif(sys.version_info < (3, 4),
-                        reason="non-contiguous memoryview not implemented in "
-                               "old Python versions")
     def test_sliced_and_non_contiguous_memoryview(self):
         buffer_obj = memoryview(b"Hello!" * 3)[2:15:2]
         self.assertEqual(pickle_depickle(buffer_obj, protocol=self.protocol),
@@ -336,9 +334,6 @@ class CloudPickleTest(unittest.TestCase):
         self.assertEqual(depickled_C.instance_of_C.it_works(), "woohoo!")
         self.assertEqual(depickled_instance.it_works(), "woohoo!")
 
-    @pytest.mark.skipif(sys.version_info >= (3, 4)
-                        and sys.version_info < (3, 4, 3),
-                        reason="subprocess has a bug in 3.4.0 to 3.4.2")
     def test_locally_defined_function_and_class(self):
         LOCAL_CONSTANT = 42
 
@@ -473,8 +468,6 @@ class CloudPickleTest(unittest.TestCase):
 
         g = pickle_depickle(F.f, protocol=self.protocol)
         self.assertEqual(g.__name__, F.f.__name__)
-        if sys.version_info[0] < 3:
-            self.assertEqual(g.im_class.__name__, F.f.im_class.__name__)
         # self.assertEqual(g(F(), 1), 2)  # still fails
 
     def test_module(self):
@@ -843,9 +836,8 @@ class CloudPickleTest(unittest.TestCase):
         assert depickled_clsdict_meth is clsdict_slotmethod
 
     @pytest.mark.skipif(
-        platform.python_implementation() == "PyPy" or
-        sys.version_info[:1] < (3,),
-        reason="No known staticmethod example in the python 2 / pypy stdlib")
+        platform.python_implementation() == "PyPy",
+        reason="No known staticmethod example in the pypy stdlib")
     def test_builtin_staticmethod(self):
         obj = "foo"  # str object
 
@@ -1740,33 +1732,6 @@ class CloudPickleTest(unittest.TestCase):
         """.format(protocol=self.protocol)
         assert_run_python_script(code)
 
-    @pytest.mark.skipif(sys.version_info >= (3, 0),
-                        reason="hardcoded pickle bytes for 2.7")
-    def test_function_pickle_compat_0_4_0(self):
-        # The result of `cloudpickle.dumps(lambda x: x)` in cloudpickle 0.4.0,
-        # Python 2.7
-        pickled = (b'\x80\x02ccloudpickle.cloudpickle\n_fill_function\nq\x00(c'
-            b'cloudpickle.cloudpickle\n_make_skel_func\nq\x01ccloudpickle.clou'
-            b'dpickle\n_builtin_type\nq\x02U\x08CodeTypeq\x03\x85q\x04Rq\x05(K'
-            b'\x01K\x01K\x01KCU\x04|\x00\x00Sq\x06N\x85q\x07)U\x01xq\x08\x85q'
-            b'\tU\x07<stdin>q\nU\x08<lambda>q\x0bK\x01U\x00q\x0c))tq\rRq\x0eJ'
-            b'\xff\xff\xff\xff}q\x0f\x87q\x10Rq\x11}q\x12N}q\x13NtR.')
-        self.assertEqual(42, cloudpickle.loads(pickled)(42))
-
-    @pytest.mark.skipif(sys.version_info >= (3, 0),
-                        reason="hardcoded pickle bytes for 2.7")
-    def test_function_pickle_compat_0_4_1(self):
-        # The result of `cloudpickle.dumps(lambda x: x)` in cloudpickle 0.4.1,
-        # Python 2.7
-        pickled = (b'\x80\x02ccloudpickle.cloudpickle\n_fill_function\nq\x00(c'
-            b'cloudpickle.cloudpickle\n_make_skel_func\nq\x01ccloudpickle.clou'
-            b'dpickle\n_builtin_type\nq\x02U\x08CodeTypeq\x03\x85q\x04Rq\x05(K'
-            b'\x01K\x01K\x01KCU\x04|\x00\x00Sq\x06N\x85q\x07)U\x01xq\x08\x85q'
-            b'\tU\x07<stdin>q\nU\x08<lambda>q\x0bK\x01U\x00q\x0c))tq\rRq\x0eJ'
-            b'\xff\xff\xff\xff}q\x0f\x87q\x10Rq\x11}q\x12N}q\x13U\x08__main__q'
-            b'\x14NtR.')
-        self.assertEqual(42, cloudpickle.loads(pickled)(42))
-
     def test_pickle_reraise(self):
         for exc_type in [Exception, ValueError, TypeError, RuntimeError]:
             obj = RaiserOnPickle(exc_type("foo"))
@@ -1822,12 +1787,7 @@ class CloudPickleTest(unittest.TestCase):
 
         self.assertEqual(f2.__doc__, f.__doc__)
 
-    @unittest.skipIf(sys.version_info < (3, 7),
-                     "This syntax won't work on py2 and pickling annotations "
-                     "isn't supported for py37 and below.")
     def test_wraps_preserves_function_annotations(self):
-        from functools import wraps
-
         def f(x):
             pass
 
@@ -1841,13 +1801,7 @@ class CloudPickleTest(unittest.TestCase):
 
         self.assertEqual(f2.__annotations__, f.__annotations__)
 
-    @unittest.skipIf(sys.version_info < (3, 7),
-                     """This syntax won't work on py2 and pickling annotations
-                     isn't supported for py37 and below.""")
     def test_type_hint(self):
-        # Try to pickle compound typing constructs. This would typically fail
-        # on Python < 3.7 (See #193)
-        import typing
         t = typing.Union[list, int]
         assert pickle_depickle(t) == t
 
@@ -1894,11 +1848,7 @@ class CloudPickleTest(unittest.TestCase):
         pickle_depickle(DataClass, protocol=self.protocol)
         assert data.x == pickle_depickle(data, protocol=self.protocol).x == 42
 
-    @unittest.skipIf(sys.version_info[:2] < (3, 5),
-                     "Only support enum pickling on Python 3.5+")
     def test_locally_defined_enum(self):
-        enum = pytest.importorskip("enum")
-
         class StringEnum(str, enum.Enum):
             """Enum when all members are also (and must be) strings"""
 
@@ -1927,10 +1877,7 @@ class CloudPickleTest(unittest.TestCase):
         green3 = pickle_depickle(Color.GREEN, protocol=self.protocol)
         assert green3 is Color.GREEN
 
-    @unittest.skipIf(sys.version_info[:2] < (3, 5),
-                     "Only support enum pickling on Python 3.5+")
     def test_locally_defined_intenum(self):
-        enum = pytest.importorskip("enum")
         # Try again with a IntEnum defined with the functional API
         DynamicColor = enum.IntEnum("Color", {"RED": 1, "GREEN": 2, "BLUE": 3})
 
@@ -1943,10 +1890,7 @@ class CloudPickleTest(unittest.TestCase):
         assert green1 is not ClonedDynamicColor.BLUE
         assert ClonedDynamicColor is DynamicColor
 
-    @unittest.skipIf(sys.version_info[:2] < (3, 5),
-                     "Only support enum pickling on Python 3.5+")
     def test_interactively_defined_enum(self):
-        pytest.importorskip("enum")
         code = """if __name__ == "__main__":
         from enum import Enum
         from testutils import subprocess_worker
@@ -2006,28 +1950,16 @@ class CloudPickleTest(unittest.TestCase):
             cloned_func = pickle_depickle(func, protocol=self.protocol)
             assert cloned_func() == "hello from a {}!".format(source)
 
-    @pytest.mark.skipif(sys.version_info[0] < 3,
-                        reason="keyword only arguments were introduced in "
-                               "python 3")
     def test_interactively_defined_func_with_keyword_only_argument(self):
         # fixes https://github.com/cloudpipe/cloudpickle/issues/263
-        # The source code of this test is bundled in a string and is ran from
-        # the __main__ module of a subprocess in order to avoid a SyntaxError
-        # in python2 when pytest imports this file, as the keyword-only syntax
-        # is python3-only.
-        code = """
-        from cloudpickle import loads, dumps
-
         def f(a, *, b=1):
             return a + b
 
-        depickled_f = loads(dumps(f, protocol={protocol}))
+        depickled_f = pickle_depickle(f, protocol=self.protocol)
 
         for func in (f, depickled_f):
             assert func(2) == 3
-            assert func.__kwdefaults__ == {{'b': 1}}
-        """.format(protocol=self.protocol)
-        assert_run_python_script(textwrap.dedent(code))
+            assert func.__kwdefaults__ == {'b': 1}
 
     @pytest.mark.skipif(not hasattr(types.CodeType, "co_posonlyargcount"),
                         reason="Requires positional-only argument syntax")
@@ -2105,11 +2037,167 @@ class CloudPickleTest(unittest.TestCase):
         reconstructed = pickle.loads(pickle_bytes, buffers=buffers)
         np.testing.assert_allclose(reconstructed.data, data_instance.data)
 
+    def test_pickle_dynamic_typevar(self):
+        T = typing.TypeVar('T')
+        depickled_T = pickle_depickle(T, protocol=self.protocol)
+        attr_list = [
+            "__name__", "__bound__", "__constraints__", "__covariant__",
+            "__contravariant__"
+        ]
+        for attr in attr_list:
+            assert getattr(T, attr) == getattr(depickled_T, attr)
+
+    def test_pickle_dynamic_typevar_tracking(self):
+        T = typing.TypeVar("T")
+        T2 = subprocess_pickle_echo(T, protocol=self.protocol)
+        assert T is T2
+
+    def test_pickle_dynamic_typevar_memoization(self):
+        T = typing.TypeVar('T')
+        depickled_T1, depickled_T2 = pickle_depickle((T, T),
+                                                     protocol=self.protocol)
+        assert depickled_T1 is depickled_T2
+
+    def test_pickle_importable_typevar(self):
+        from .mypkg import T
+        T1 = pickle_depickle(T, protocol=self.protocol)
+        assert T1 is T
+
+        # Standard Library TypeVar
+        from typing import AnyStr
+        assert AnyStr is pickle_depickle(AnyStr, protocol=self.protocol)
+
+    def test_generic_type(self):
+        T = typing.TypeVar('T')
+
+        class C(typing.Generic[T]):
+            pass
+
+        assert pickle_depickle(C, protocol=self.protocol) is C
+
+        # Identity is not part of the typing contract: only test for
+        # equality instead.
+        assert pickle_depickle(C[int], protocol=self.protocol) == C[int]
+
+        with subprocess_worker(protocol=self.protocol) as worker:
+
+            def check_generic(generic, origin, type_value):
+                assert generic.__origin__ is origin
+                assert len(generic.__args__) == 1
+                assert generic.__args__[0] is type_value
+
+                assert len(origin.__orig_bases__) == 1
+                ob = origin.__orig_bases__[0]
+                assert ob.__origin__ is typing.Generic
+                assert len(ob.__parameters__) == 1
+
+                return "ok"
+
+            assert check_generic(C[int], C, int) == "ok"
+            assert worker.run(check_generic, C[int], C, int) == "ok"
+
+    def test_locally_defined_class_with_type_hints(self):
+        with subprocess_worker(protocol=self.protocol) as worker:
+            for type_ in _all_types_to_test():
+                class MyClass:
+                    def method(self, arg: type_) -> type_:
+                        return arg
+                MyClass.__annotations__ = {'attribute': type_}
+
+                def check_annotations(obj, expected_type):
+                    assert obj.__annotations__["attribute"] == expected_type
+                    assert obj.method.__annotations__["arg"] == expected_type
+                    assert (
+                        obj.method.__annotations__["return"] == expected_type
+                    )
+                    return "ok"
+
+                obj = MyClass()
+                assert check_annotations(obj, type_) == "ok"
+                assert worker.run(check_annotations, obj, type_) == "ok"
+
+    def test_generic_extensions(self):
+        typing_extensions = pytest.importorskip('typing_extensions')
+
+        objs = [
+            typing_extensions.Literal,
+            typing_extensions.Final,
+            typing_extensions.Literal['a'],
+            typing_extensions.Final[int],
+        ]
+
+        for obj in objs:
+            depickled_obj = pickle_depickle(obj, protocol=self.protocol)
+            assert depickled_obj == obj
+
+    def test_class_annotations(self):
+        class C:
+            pass
+        C.__annotations__ = {'a': int}
+
+        C1 = pickle_depickle(C, protocol=self.protocol)
+        assert C1.__annotations__ == C.__annotations__
+
+    def test_function_annotations(self):
+        def f(a: int) -> str:
+            pass
+
+        f1 = pickle_depickle(f, protocol=self.protocol)
+        assert f1.__annotations__ == f.__annotations__
+
 
 class Protocol2CloudPickleTest(CloudPickleTest):
 
     protocol = 2
 
 
+def test_lookup_module_and_qualname_dynamic_typevar():
+    T = typing.TypeVar('T')
+    module_and_name = _lookup_module_and_qualname(T, name=T.__name__)
+    assert module_and_name is None
+
+
+def test_lookup_module_and_qualname_importable_typevar():
+    from . import mypkg
+    T = mypkg.T
+    module_and_name = _lookup_module_and_qualname(T, name=T.__name__)
+    assert module_and_name is not None
+    module, name = module_and_name
+    assert module is mypkg
+    assert name == 'T'
+
+
+def test_lookup_module_and_qualname_stdlib_typevar():
+    module_and_name = _lookup_module_and_qualname(typing.AnyStr,
+                                                  name=typing.AnyStr.__name__)
+    assert module_and_name is not None
+    module, name = module_and_name
+    assert module is typing
+    assert name == 'AnyStr'
+
+
+def _all_types_to_test():
+    T = typing.TypeVar('T')
+
+    class C(typing.Generic[T]):
+        pass
+
+    return [
+        C, C[int],
+        T, typing.Any, typing.NoReturn, typing.Optional,
+        typing.Generic, typing.Union, typing.ClassVar,
+        typing.Optional[int],
+        typing.Generic[T],
+        typing.Callable[[int], typing.Any],
+        typing.Callable[..., typing.Any],
+        typing.Callable[[], typing.Any],
+        typing.Tuple[int, ...],
+        typing.Tuple[int, C[int]],
+        typing.ClassVar[C[int]],
+        typing.List[int],
+        typing.Dict[int, str],
+    ]
+
+
 if __name__ == '__main__':
     unittest.main()
